"""A parser for HCL2 implemented using the Lark parser"""
from __future__ import annotations

from importlib import resources
from typing import Any

from lark import Lark

from hcl2.transformer import DictTransformer

LARK_GRAMMAR = resources.read_text(__package__, "hcl2.lark")


def strip_line_comment(line: str) -> tuple[str, str, str] | tuple[str, None, None]:
    """
    Finds the start of a comment in the line, if any, and returns the line
    up to the comment, the token that started the comment (#, //, or /*),
    and the line after the comment token
    """
    comment_tokens = ['#', '//', '/*']

    # manual iteration; trying to avoid a bunch of repeated "in" searches
    index = 0
    while index < len(line):
        for token in comment_tokens:
            if index > len(line) - len(token):
                continue
            if line[index:index + len(token)] == token and \
                    line[0:index].replace('\\"', '').count('"') % 2 == 0:
                # we are not in a string, so this marks the start of a comment
                return line[0:index], token, line[index + len(token):]
        index += 1

    return line, None, None


class Hcl2:
    """Wrapper class for Lark"""

    lark_parser = Lark(grammar=LARK_GRAMMAR, parser="lalr", propagate_positions=True, cache=True)

    def parse(self, text: str) -> dict[str, list[dict[str, Any]]]:
        """Parses a HCL file and returns a dict"""

        tree = Hcl2.lark_parser.parse(text)
        return DictTransformer().transform(tree)


hcl2 = Hcl2()



"""The API that will be exposed to users of this package"""
from __future__ import annotations

import re
from typing import TextIO, Any

from hcl2.parser import hcl2, strip_line_comment


def load(file: TextIO) -> dict[str, list[dict[str, Any]]]:
    """Load a HCL2 file"""
    return loads(file.read())


def loads(text: str) -> dict[str, list[dict[str, Any]]]:
    """Load HCL2 from a string"""
    # append new line as a workaround for https://github.com/lark-parser/lark/issues/237
    # Lark doesn't support a EOF token so our grammar can't look for "new line or end of file"
    # This means that all blocks must end in a new line even if the file ends
    # Append a new line as a temporary fix

    # Avoid catastrophic backtracking on missing quote marks
    multi_line_string_denominator = ""

    # this will be a mostly-complete multi-line comment approach.
    # It will pick up lines that include a comment start (/*),
    # and it will check whether the comment ends on the same line.
    # But it will not handle weird things like code that
    # starts on the same line after closing a multi line comment:
    # a = 123 /* ....      <- this will get checked
    # ....
    # ... /* a = 123     <- this will not
    in_multi_line_comment = False
    found_multi_line_comment_start = False
    found_multi_line_interpolation_start = False

    for line in text.split('\n'):
        comment = None
        if not multi_line_string_denominator and \
                not in_multi_line_comment and '/*' in line:
            line, token, comment = strip_line_comment(line)
            if token == '/*':
                # handle a comment that includes a /* (e.g., `... # ... /*`
                # which is not a multi-line comment
                # first we will complete this loop and consider any part of the line that still remains
                found_multi_line_comment_start = True
        elif not multi_line_string_denominator and in_multi_line_comment and '*/' in line:
            in_multi_line_comment = False
            continue

        if not in_multi_line_comment and '= <<' in line:
            multi_line_string_denominator = line.split('= <<')[1]
        elif not in_multi_line_comment and multi_line_string_denominator \
                and re.match(rf'\s*{multi_line_string_denominator}', line):
            multi_line_string_denominator = ""
        elif not in_multi_line_comment and multi_line_string_denominator == "" \
                and line.replace('\\"', '').count('"') % 2 != 0:
            # it's possible that the unclosed quote is in a comment, so double check
            # (but we don't want to do this for every line if it's not necessary)
            stripped = strip_line_comment(line)[0] if not comment else line  # maybe we already stripped it
            if stripped.replace('\\"', '').count('"') % 2 != 0:
                if (
                    "${" in stripped
                    and "}" not in stripped
                    and stripped.index('"') < stripped.index("${")
                ):
                    # this seems to be a multiline interpolation string
                    found_multi_line_interpolation_start = True
                    continue
                if (
                    found_multi_line_interpolation_start
                    and "}" in stripped
                    and stripped.index("}") < stripped.index('"')
                ):
                    # found also the closing end
                    found_multi_line_interpolation_start = False
                    continue

                # the stripped off comment is still an unclosed string, so now we have a real error
                raise ValueError(f"Line has unclosed quote marks: {line}")

        if found_multi_line_comment_start:
            found_multi_line_comment_start = False
            in_multi_line_comment = comment is not None and '*/' not in comment
            # check whether this "multiline" comment closed on the same line

    return hcl2.parse(text + "\n")



"""A Lark Transformer for transforming a Lark parse tree into a Python dict"""
from __future__ import annotations

import re
import sys
from typing import Any, TYPE_CHECKING

from lark import Transformer, Token
from lark.visitors import v_args

if TYPE_CHECKING:
    from lark.tree import Meta

HEREDOC_PATTERN = re.compile(r"<<([a-zA-Z][a-zA-Z0-9._-]+)\n(([^\n]|\n)*)\n\s*\1", re.S)
HEREDOC_TRIM_PATTERN = re.compile(r"<<-([a-zA-Z][a-zA-Z0-9._-]+)\n(([^\n]|\n)*)\n\s*\1", re.S)

START_LINE = "__start_line__"
END_LINE = "__end_line__"

NO_BLOCK_LABEL_TYPES = {"locals", "terraform"}
ONE_BLOCK_LABEL_TYPES = {"module", "provider", "variable"}
TWO_BLOCK_LABEL_TYPES = {"data", "resource"}


# pylint: disable=missing-docstring,unused-argument
class DictTransformer(Transformer[Token, "dict[str, list[dict[str, Any]]]"]):
    def float_lit(self, args: list) -> float:
        return float("".join([str(arg) for arg in args]))

    def int_lit(self, args: list) -> int:
        return int("".join([str(arg) for arg in args]))

    def expr_term(self, args: list) -> Any:
        args = self.strip_new_line_tokens(args)

        #
        if args[0] == "true":
            return True
        if args[0] == "false":
            return False
        if args[0] == "null":
            return None

        # if the expression starts with a paren then unwrap it
        if args[0] == "(":
            return args[1]
        # otherwise return the value itself
        return args[0]

    def index_expr_term(self, args: list) -> str:
        args = self.strip_new_line_tokens(args)
        return f"{str(args[0])}{str(args[1])}"

    def index(self, args: list) -> str:
        args = self.strip_new_line_tokens(args)
        return f"[{str(args[0])}]"

    def get_attr_expr_term(self, args: list) -> str:
        return f"{str(args[0])}.{str(args[1])}"

    def attr_splat_expr_term(self, args: list) -> str:
        return f"{args[0]}.*.{args[1]}"

    def full_splat_expr_term(self, args: list) -> str:
        return f"{args[0]}[*].{args[1]}"

    def tuple(self, args: list) -> list:
        return [self.to_string_dollar(arg) for arg in self.strip_new_line_tokens(args)]

    def object_elem(self, args: list) -> dict[str, Any]:
        # This returns a dict with a single key/value pair to make it easier to merge these
        # into a bigger dict that is returned by the "object" function
        key = self.strip_quotes(args[0])
        value = self.to_string_dollar(args[1])

        return {
            key: value,
        }

    def object(self, args: list) -> dict:
        args = self.strip_new_line_tokens(args)
        result: dict[str, Any] = {}
        for arg in args:
            result.update(arg)
        return result

    def function_call(self, args: list) -> str:
        args = self.strip_new_line_tokens(args)
        args_str = ""
        if len(args) > 1:
            args_str = ",".join([str(arg) for arg in args[1]])
        return f"{str(args[0])}({args_str})"

    def arguments(self, args: list) -> list:
        return args

    @v_args(meta=True)
    def block(self, meta: Meta, args: list) -> dict[str, Any]:
        args = self.strip_new_line_tokens(args)

        # if the last token is a string instead of an object then the block is empty
        # such as 'foo "bar" "baz" {}'
        # in that case append an empty object
        if isinstance(args[-1], str):
            args.append({})

        result: dict[str, Any] = {}
        current_level = result
        for arg in args[0:-2]:
            current_level[self.strip_quotes(arg)] = {}
            current_level = current_level[self.strip_quotes(arg)]

        current_level[self.strip_quotes(args[-2])] = args[-1]

        if args[0] in TWO_BLOCK_LABEL_TYPES and isinstance(args[1], str) and isinstance(args[2], str):
            label_1 = self.strip_quotes(args[1])
            label_2 = self.strip_quotes(args[2])
            result[args[0]][label_1][label_2][START_LINE] = meta.line
            result[args[0]][label_1][label_2][END_LINE] = meta.end_line

        if args[0] in ONE_BLOCK_LABEL_TYPES and isinstance(args[1], str):
            label_1 = self.strip_quotes(args[1])
            result[args[0]][label_1][START_LINE] = meta.line
            result[args[0]][label_1][END_LINE] = meta.end_line

        if args[0] in NO_BLOCK_LABEL_TYPES:
            result[args[0]][START_LINE] = meta.line
            result[args[0]][END_LINE] = meta.end_line

        return result

    def attribute(self, args: list) -> dict[str, Any]:
        key = str(args[0])
        if key.startswith('"') and key.endswith('"'):
            key = key[1:-1]
        value = self.to_string_dollar(args[1])

        return {
            key: value,
        }

    def conditional(self, args: list) -> str:
        args = self.strip_new_line_tokens(args)
        return f"{args[0]} ? {args[1]} : {args[2]}"

    def binary_op(self, args: list) -> str:
        return " ".join([str(arg) for arg in args])

    def unary_op(self, args: list) -> str:
        return "".join([str(arg) for arg in args])

    def binary_term(self, args: list) -> str:
        args = self.strip_new_line_tokens(args)
        return " ".join([str(arg) for arg in args])

    def body(self, args: list) -> dict[str, list]:
        # A body can have multiple attributes with the same name
        # For example multiple Statement attributes in a IAM resource body
        # So This returns a dict of attribute names to lists
        # The attribute values will always be lists even if they aren't repeated
        # and only contain a single entry
        args = self.strip_new_line_tokens(args)
        result: dict[str, Any] = {}
        for arg in args:
            for key, value in arg.items():
                key = str(key)
                if key not in result:
                    result[key] = [value]
                else:
                    if isinstance(result[key], list):
                        if isinstance(value, list):
                            result[key].extend(value)
                        else:
                            result[key].append(value)
                    else:
                        result[key] = [result[key], value]

        return result

    def start(self, args: list) -> dict[str, list[dict[str, Any]]]:
        args = self.strip_new_line_tokens(args)
        return args[0]

    def binary_operator(self, args: list) -> str:
        return str(args[0])

    def heredoc_template(self, args: list) -> str:
        match = HEREDOC_PATTERN.match(str(args[0]))
        if not match:
            raise RuntimeError(f"Invalid Heredoc token: {args[0]}")
        return f'"{match.group(2)}"'

    def heredoc_template_trim(self, args: list) -> str:
        # See https://github.com/hashicorp/hcl2/blob/master/hcl/hclsyntax/spec.md#template-expressions
        # This is a special version of heredocs that are declared with "<<-"
        # This will calculate the minimum number of leading spaces in each line of a heredoc
        # and then remove that number of spaces from each line
        match = HEREDOC_TRIM_PATTERN.match(str(args[0]))
        if not match:
            raise RuntimeError(f"Invalid Heredoc token: {args[0]}")

        text = match.group(2)
        lines = text.split("\n")

        # calculate the min number of leading spaces in each line
        min_spaces = sys.maxsize
        for line in lines:
            leading_spaces = len(line) - len(line.lstrip(" "))
            min_spaces = min(min_spaces, leading_spaces)

        # trim off that number of leading spaces from each line
        lines = [line[min_spaces:] for line in lines]

        return '"{}"'.format("\n".join(lines))

    def for_tuple_expr(self, args: list) -> str:
        args = self.strip_new_line_tokens(args)
        for_expr = " ".join([str(arg) for arg in args[1:-1]])
        return f"[{for_expr}]"

    def for_intro(self, args: list) -> str:
        args = self.strip_new_line_tokens(args)
        return " ".join([str(arg) for arg in args])

    def for_cond(self, args: list) -> str:
        args = self.strip_new_line_tokens(args)
        return " ".join([str(arg) for arg in args])

    def for_object_expr(self, args: list) -> str:
        args = self.strip_new_line_tokens(args)
        for_expr = " ".join([str(arg) for arg in args[1:-1]])
        return f"{{{for_expr}}}"

    def strip_new_line_tokens(self, args: list[Any]) -> list[Any]:
        """
        Remove new line and Discard tokens.
        The parser will sometimes include these in the tree so we need to strip them out here
        """
        return [arg for arg in args if arg != "\n"]

    def to_string_dollar(self, value: Any) -> Any:
        """Wrap a string in ${ and }"""
        if isinstance(value, str):
            if value.startswith('"') and value.endswith('"'):
                return str(value)[1:-1]
            return f"${{{value}}}"
        return value

    def strip_quotes(self, value: Any) -> Any:
        """Remove quote characters from the start and end of a string"""
        if isinstance(value, str):
            if value.startswith('"') and value.endswith('"'):
                return str(value)[1:-1]
        return value

    def identifier(self, value: Any) -> Any:
        # Making identifier a token by capitalizing it to IDENTIFIER
        # seems to return a token object instead of the str
        # So treat it like a regular rule
        # In this case we just convert the whole thing to a string
        return str(value[0])
